//
// XrmAudioServer
//
// Handles AudioWebSocket communication events and frames processing.
//
// Depends on:
//  * xrm-globals.js
//  * xrm-websocket.js
//  * qos-stats.js
//

//set up webaudio
var audioContext;

try {
    window.AudioContext = window.AudioContext || window.webkitAudioContext;
    audioContext = new AudioContext({
        latencyHint: "interactive"
    });
} catch(e) {
    try {
        audioContext = new AudioContext();
    } catch (e) {
        alert('Web Audio API is not supported in this browser');
    }
}

//
// Audio states.
//
var XRM_AUDIO_SESSION_CLOSED = 0; // audio is not initialized
var XRM_AUDIO_SESSION_OPENED = 1; // websocket connected and client is waiting for acknowledgement
var XRM_AUDIO_SESSION_CONNECTED = 2; // websocket connected, client is acknowledged, and server is waiting for a command ('play')
var XRM_AUDIO_SESSION_ACTIVATED = 3; // audio session is playing over websocket

function XrmAudioServer(dataWS, logger) {
    this.dataWS = dataWS;

    this.audioWS;

    // The state of the audio session.
    this.sessionState = XRM_AUDIO_SESSION_CLOSED;
    this.cachedSessionState = null;

    this.audioEnabled = false;

    this.stats = new Stats();

    this.sampleRate = 44100;
    this.startDelay = 0.050;
    this.bufferTime = 0.020;
    this.sampleRateAdjust = 1.01;
    this.channels = 2;
    this.bitsPerSample = 16;
    this.threshold = 0.150;
    this.samplesPerFrame = 512;

    // static variables of playBufd
    this.copyBufferLength = this.sampleRate;
    this.nextStartTime = 0;
    this.counter = 0;
    this.lBuf = new Float32Array(this.copyBufferLength);
    this.rBuf = new Float32Array(this.copyBufferLength);
    this.bufPtr = 0;

    var timestamps = new tStamps();

    this.setIsEnabled = function(val) {
        this.audioEnabled = val;
    }.bind(this);

    this.isEnabled = function() {
        return this.audioEnabled === true;
    }.bind(this);

    this.setSessionState = function(v) {
        this.sessionState = v;
    }.bind(this);

    this.getSessionState = function() {
        return this.sessionState;
    }

    this.cacheSessionState = function() {
        this.cachedSessionState = this.getSessionState();
    }

    this.resetCachedSessionState = function() {
        this.cachedSessionState = null;
    }

    this.getCachedSessionState = function() {
        return this.cachedSessionState;
    }

    this.isSessionClosed = function() {
        return this.sessionState === XRM_AUDIO_SESSION_CLOSED;
    }.bind(this);

    this.isSessionOpened = function() {
        return this.sessionState === XRM_AUDIO_SESSION_OPENED;
    }.bind(this);

    this.isSessionConnected = function() {
        return this.sessionState === XRM_AUDIO_SESSION_CONNECTED;
    }.bind(this);

    this.isSessionActive = function() {
        return this.sessionState === XRM_AUDIO_SESSION_ACTIVATED;
    }.bind(this);

    this.sendPlay = function() {
        if (this.isSessionActive() || !this.isEnabled())
            return;

        var obj = {};
        obj[XRM_MSG_TYPE] = XRM_MSGTYPE_CLNT_START_AS;
        this.dataWS.send(JSON.stringify(obj));
        logger.debug("Audio: Sent play audio source message");
        this.setSessionState(XRM_AUDIO_SESSION_ACTIVATED);
    }.bind(this);

    this.stopAudioStream = function() {
        if (this.isSessionClosed() || this.isSessionConnected() || !this.isEnabled())
            return;

        var obj = {};
        obj[XRM_MSG_TYPE] = XRM_MSGTYPE_CLNT_STOP_AS;
        this.dataWS.send(JSON.stringify(obj));
        logger.debug("Audio: Sent stop audio source message");

        this.setSessionState(XRM_AUDIO_SESSION_CONNECTED);
    }.bind(this);

    this.sendSetup = function() {
        if (!this.isEnabled())
            return;

        var obj = {};
        obj[XRM_MSG_TYPE] = XRM_MSGTYPE_CLNT_AUDIO_SETUP;
        this.dataWS.send(JSON.stringify(obj));
        logger.debug("Audio: Sent setup audio source message");
    }.bind(this);

    this.parseAudioParams = function(msg) {
        if (!this.isEnabled())
            return;

        this.setOutputFormat(msg[XRM_AUDIO_PARAM_OUTPUT_FORMAT]);
        this.sampleRate = msg[XRM_AUDIO_PARAM_SAMPLE_RATE];
        this.startDelay = msg[XRM_AUDIO_PARAM_START_DELAY];
        this.bufferTime = msg[XRM_AUDIO_PARAM_BUFFER_TIME];
        this.sampleRateAdjust = msg[XRM_AUDIO_PARAM_SAMPLERATEDADJUST];
        this.channels = msg[XRM_AUDIO_PARAM_CHANNELS];
        this.bitsPerSample = msg[XRM_AUDIO_PARAM_BITSPERSAMPLE];
        this.threshold = msg[XRM_AUDIO_PARAM_THRESHOLD];
        this.samplesPerFrame = msg[XRM_AUDIO_PARAM_SAMPLES_PER_FRAME];

        this.copyBufferLength = this.sampleRate;
        this.nextStartTime = 0;
        this.counter = 0;
        this.lBuf = new Float32Array(this.copyBufferLength);
        this.rBuf = new Float32Array(this.copyBufferLength);
        this.bufPtr = 0;

    }.bind(this);

    this.playAudioStream = function(clientID) {
        if (!this.isEnabled())
            return;

        this.clientID = clientID;

        // WebSocket is already connected and acknowledged.  Server is ready for play command.
        if (this.isSessionConnected()) {
            this.sendPlay();
            return;
        }

        // WebSocket is connected and must wait for server acknowledgement, or we just received an format update
        if (this.isSessionOpened() || this.isSessionActive()) {
            return;
        }

        //
        // Websocket not connected - make a new one.
        //
        this.audioWS = new XrmWebSocket();
        this.audioWS.open(XrmWebSocket.defaultProtocol() + location.host + "/ws/desktop/audio/" + this.clientID + "/0");
        this.audioWS.setBinaryType("arraybuffer");     // make sure data arrives as binary data
        this.setSessionState(XRM_AUDIO_SESSION_OPENED);
        this.nextStartTime = 0;
        logger.log("Connected audio WS");

        this.audioWS.onclose = function() {
            this.setSessionState(XRM_AUDIO_SESSION_CLOSED);
        }.bind(this);

        this.audioWS.onmessage = function(evt) {
            try {
                var msg = evt.data;

                this.stats.frameCtr++;

                // logger.debug("Received frame = " + this.stats.frameCtr + ", size " + msg.byteLength);

                try {
                    timestamps.Entry = Date.now();
                    playBuffer(msg);
                    timestamps.Exit = Date.now();

                    if (timestamps.Exit + 1000 > this.stats.frameCtrIntStart) {
                        this.stats.frameCtrPrevInt = this.stats.frameCtrCurrInt;
                        this.stats.frameCtrCurrInt = 0;
                        while (timestamps.Exit + 1000 > this.stats.frameCtrIntStart) {
                            this.stats.frameCtrIntStart += 1000;
                        }
                    }
                    this.stats.frameCtrCurrInt++;
                } catch(err) {
                    logger.warn("Decoding error " + err  + ", frame size " + msg.byteLength);
                }
            } catch(err) {
                logger.error("Error, msg " + err);
            }
        }.bind(this);
    }.bind(this);

    this.handleAudioSourceParam = function(jsonMessage, nativeDecoderEnabled, serverFeatures, webGLSupported) {
        logger.log("Received AudioSource Parameter");
    }.bind(this);

    // Translate buffers with 24-bit signed ints (little endian) into a 32 bit float array.
    // NOTE: data.length % BYTES_PER_ELEMENT must == 0
    this.__translateFormat24 = function(buffer) {
        const BYTES_PER_ELEMENT = 3;
        var data = new Uint8Array(buffer);
        var float32 = new Float32Array(data.length / BYTES_PER_ELEMENT);

        var j = 0;
        const maxv = this.maxValue;
        for (var i = 0; i < data.length; i += BYTES_PER_ELEMENT) {
            // Below is faster equivalent of: val = ((data[i + 2] << 24) | (data[i + 1] << 16) | (data[i] << 8)) >> 8;
            var val = ((data[i + 2] << 24) >> 8);
            val = val | (data[i + 1] << 8) | (data[i]);

            float32[j] = val / maxv;
            j++;
        }
        return float32.buffer;
    }

    // Translates buffers with 8, 16, 32, and 64 bit signed ints (little endian) into a 32 bit float array.
    // NOTE: data.length % BYTES_PER_ELEMENT must == 0
    this.__translateFormat = function(buffer) {
        var data = new this.TypedArray(buffer);
        var float32 = new Float32Array(data.length);

        const maxv = this.maxValue;
        for (var i = 0; i < data.length; i++) {
            float32[i] = data[i] / maxv;
        }
        return float32.buffer;
    };

    this.setOutputFormat = function(format) {
        this.outputFormat = format;
        this.translateFormat = this.__translateFormat;

        switch(this.outputFormat) {
            case "s16le":
                this.maxValue = 32768;
                this.TypedArray = Int16Array;
                break;
            case "s8":
                this.maxValue = 128;
                this.TypedArray = Int8Array;
                break;
            case "s24le":
                this.maxValue = 8388608;
                this.TypedArray = undefined;
                this.translateFormat = this.__translateFormat24;
                break;
            case "s32le":
                this.maxValue = 2147483648;
                this.TypedArray = Int32Array;
                break;
            case "f32le":
                this.maxValue = 1;
                this.TypedArray = Float32Array;
                break;
            case "f64le":
                this.maxValue = 0.5;
                this.TypedArray = Float64Array;
                break;
            default:
                this.maxValue = 32768;
                this.TypedArray = Int16Array;
                break;
        };
    }

    // NOTE: setOutputFormat MUST be called before this.
    // fills L/R buffers with chunk and plays them when L/R buffers are full
    var playBuffer = function(data) {
        var chunk = data;

        if (this.outputFormat !== "f32le")
            chunk = this.translateFormat(data);

        // console.log("this.sampleRate= " + this.sampleRate +
        //             " this.startDelay= " + this.startDelay +
        //             " this.bufferTime= " + this.bufferTime +
        //             " this.sampleRateAdjust=" + this.sampleRateAdjust +
        //             " this.threshold= " + this.threshold);

        const chunkBuf = new DataView(chunk);
        //const startDelay = 0.025;
        //const bufferTime = 0.075; //length of buffer in ms
        const playTrigger = this.copyBufferLength * this.bufferTime;
        // number of samples per canal in chunk
        const sampleSize = Float32Array.BYTES_PER_ELEMENT;
        const nSamples = chunkBuf.byteLength / sampleSize / 2;

        this.counter++;

        // add incoming audio frame to l/rBuf
        // convert incoming chunk into 32 bit float array as needed by webaudio
        for (var i = 0; i < nSamples; i++) {
            this.lBuf[this.bufPtr + i] = chunkBuf.getFloat32(2 * i * sampleSize, true);
            this.rBuf[this.bufPtr + i] = chunkBuf.getFloat32((2 * i + 1) * sampleSize, true);
        }
        this.bufPtr += nSamples;

        // l/rBuf long enough to play ?
        if (this.bufPtr > playTrigger) {
            // create webaudio buffer and populate with copyBuffers
            var audioBuffer = audioContext.createBuffer(2, this.bufPtr, this.sampleRate);
            audioBuffer.getChannelData(0).set(this.lBuf.slice(0, this.bufPtr));
            audioBuffer.getChannelData(1).set(this.rBuf.slice(0, this.bufPtr));

            var source = audioContext.createBufferSource(); // creates a sound source
            source.buffer = audioBuffer;
            source.connect(audioContext.destination); // connect the source to the context's destination (the speakers)

            // if first call of playBufd, delay play of chunk a bit
            if (this.nextStartTime == 0) {
                this.nextStartTime = audioContext.currentTime + this.startDelay;
            }

            // adjust sampleRate if audio gets behind or ahead
            // use times from audio frame played last
            var sampleRate = this.sampleRate;
            var tDelay = this.nextStartTime - audioContext.currentTime;
            // if time to play audio buffer is delayed, adjust the samplerate to play buffer a bit faster
            if (tDelay > this.threshold) {
                sampleRate = sampleRate * this.sampleRateAdjust;
                //adjust playback rate
                source.playbackRate.value = this.sampleRateAdjust;
            }

            // in case client recieved packages much later (i.e. pulse audio behavior when volume changes)
            // reset nextStartTime
            if (tDelay < 0.0) {
                this.nextStartTime = audioContext.currentTime + this.startDelay;
            }

            source.start(this.nextStartTime);

            // console.log(":counter= " + this.counter +
            //             " :t= " + audioContext.currentTime.toFixed(3) +
            //             " :td= " + (audioBuffer.length / sampleRate).toFixed(3) +
            //             " :tnext= " + this.nextStartTime.toFixed(3) +
            //             " :tDelay=" + tDelay.toFixed(3) +
            //             " :samples= " + nSamples +
            //             " :sampleRate= " + sampleRate + "\n");

            //console.log("this.lBuf = " + this.lBuf.slice(0,16));
            //console.log(""); //insert new line

            // set startTime for the next audio frame
            var delta = audioBuffer.length / sampleRate;
            if (tDelay > 3 * this.threshold) {
                delta -= 3 * this.threshold;
            }
            this.nextStartTime += delta;

            // reset buffer pointer
            this.bufPtr = 0;
        }
    }.bind(this);

}

//
// Safari requires that we play a sound in an "onclick" event handler
// to play audio for the lifetime of the audioContext.  This plays an
// empty sound.
//
// NOTE: If the audioContext is lost / reset any new ones will not
// retain permission.
//
function playNullSound() {
    const buffer = audioContext.createBuffer(1, 1, audioContext.sampleRate);
    var sample = audioContext.createBufferSource();
    sample.buffer = buffer;
    sample.connect(audioContext.destination);
    sample.start();
}
