//
// XrmVideoServer
//
// Handles VideoWebSocket communication events and frames processing.
// Requires a data websocket to handle control messages.
//
// Depends on:
//  * xrm-globals.js
//  * xrm-websocket.js
//  * system-alert.js
//  * native-client.js
//  * qos-stats.js
//

function VideoQualitySetting(name, code, isScalable) {
    this.name = name;
    this.code = code;
    this.isScalable = isScalable;
}

// code: the x11-remote code
var VIDEO_QUALITY_SETTINGS = [
    new VideoQualitySetting("lossy", "0,0,8", true),
    new VideoQualitySetting("visuallyLossless", "6,6,8", false),
    new VideoQualitySetting("trulyLossless", "9,125,8", false),
];

// Allows NativeCanvas to be seen (rather than being blocked)
var enableTransparency = function() {
    logger.log("Enabling transparency");
    $("canvas").addClass("transparent-background");
    $("body").addClass("transparent-background");
};

//
// Video session states.
//
var XRM_VIDEO_SESSION_CLOSED = 0; // video session is not initialized
var XRM_VIDEO_SESSION_OPENED = 1; // websocket connected and client is waiting for acknowledgement
var XRM_VIDEO_SESSION_CONNECTED = 2; // websocket connected, client is acknowledged, and server is waiting for a command ('play')
var XRM_VIDEO_SESSION_ACTIVATED = 3; // video session is playing over websocket

function XrmVideoServer(canvas, dataWS, nativeClientEnabled, logger)
{
    this.dataWS = dataWS;

    this.videoWS;

    // The state of the video session.
    this.sessionState = XRM_VIDEO_SESSION_CLOSED;
    // The number of times this has been in the XRM_VIDEO_SESSION_ACTIVATED state
    this.sessionCounter = 0;

    this.currentEncoding = VIDEO_QUALITY_SETTINGS[0].code;

    this.stats = new Stats();

    this.canvas = canvas;
    this.reportInterval = -1;
    this.lastSeqNoReported = -1;

    this.avcPlayer = null;

    var timestamps = new tStamps();

    this.setSessionState = function(v) {
        if (v == XRM_VIDEO_SESSION_ACTIVATED) {
            this.sessionCounter += 1;
        }
        this.sessionState = v;
    }.bind(this);

    this.getSessionState = function() {
        return this.sessionState;
    }

    this.isSessionClosed = function() {
        return this.sessionState === XRM_VIDEO_SESSION_CLOSED;
    }.bind(this);

    this.isSessionOpened = function() {
        return this.sessionState === XRM_VIDEO_SESSION_OPENED;
    }.bind(this);

    this.isSessionConnected = function() {
        return this.sessionState === XRM_VIDEO_SESSION_CONNECTED;
    }.bind(this);

    this.isSessionActive = function() {
        return this.sessionState === XRM_VIDEO_SESSION_ACTIVATED;
    }.bind(this);

    this.sendPlay = function() {
        var obj = {};
        obj[XRM_MSG_TYPE] = XRM_MSGTYPE_CLNT_START_VS;
        obj[XRM_MSG_SCREEN_ID] = this.screenID;
        this.dataWS.send(JSON.stringify(obj));
        logger.debug("Sent play video source message");
        this.setSessionState(XRM_VIDEO_SESSION_ACTIVATED);
    }.bind(this);

    this.sendSetup = function(screenID, encoding) {
        this.screenID = screenID;
        var obj = {};
        obj[XRM_MSG_TYPE] = XRM_MSGTYPE_CLNT_VIDEO_SETUP;
        obj[XRM_MSG_SCREEN_ID] = this.screenID;
        obj[XRM_VCONF] = encoding;
        this.dataWS.send(JSON.stringify(obj));
        logger.debug("Sent setup video source message");
    }.bind(this);

    // use "0,0,8" for lossy H264 (yuv420), "9,125,8" for lossless H264 (RGBp)
    // But make sure, we are running native client !!!
    this.sendEncodingChangeRequest = function(encoding) {
        var obj = {};
        obj[XRM_MSG_TYPE] = XRM_MSGTYPE_CLNT_PARAM_CHANGE;
        obj[XRM_MSG_SCREEN_ID] = this.screenID;
        obj[XRM_VCONF] = encoding;
        this.dataWS.send(JSON.stringify(obj));
        logger.log("Sent request to change encoding to: " + encoding);
    }.bind(this);

    this.openVideoStream = function(clientID, screenID) {
        this.clientID = clientID;
        this.screenID = screenID;
        // WebSocket is connected and acknowledged.  Server is ready for play command.
        if (this.isSessionConnected()) {
            this.sendPlay();
            return;
        }

        // WebSocket is connected and must wait for server acknowledgement
        if (this.isSessionOpened()) {
            return;
        }

        // Session is aktive (e.g during swtich of VS), signal server
        if (this.isSessionActive()) {
            this.sendPlay();
            return;
        }

        //
        // Websocket not connected - make a new one.
        //
        this.videoWS = new XrmWebSocket();
        this.videoWS.open(XrmWebSocket.defaultProtocol() + location.host + "/ws/desktop/screen/" + this.clientID + "/" + this.screenID);
        //make sure data arrives as binary data
        this.videoWS.setBinaryType("arraybuffer");
        this.setSessionState(XRM_VIDEO_SESSION_OPENED);
        logger.debug("Connected video WS");

        this.videoWS.onclose = function() {
            this.setSessionState(XRM_VIDEO_SESSION_CLOSED);
        }.bind(this);

        this.videoWS.onmessage = function(evt) {
            try {
                var msg = evt.data;

                this.stats.frameCtr++;

                // logger.trace("Received frame = " + this.stats.frameCtr + ", size " + msg.byteLength);

                if (this.stats.frameCtr >= this.nextFrameNo) {
                    if (this.stats.frameCtr === this.nextFrameNo) {
                        logger.debug("Resync: Frame " + this.stats.frameCtr + " size " + msg.byteLength);
                    }

                    try {
                        var h264Frame = new Uint8Array(msg);
                        timestamps.Entry = Date.now();
                        this.avcPlayer.decode(h264Frame);
                        timestamps.Exit = Date.now();
                        if (timestamps.Exit + 1000 > this.stats.frameCtrIntStart)
                        {
                            this.stats.frameCtrPrevInt = this.stats.frameCtrCurrInt;
                            this.stats.frameCtrCurrInt = 0;
                            while (timestamps.Exit + 1000 > this.stats.frameCtrIntStart)
                                this.stats.frameCtrIntStart += 1000;
                        }
                        this.stats.frameCtrCurrInt++;
                    } catch(err) {
                        logger.warn("Decoding error, this.canvas size = " + this.canvas.width + "x" + this.canvas.height + ", err msg " + err  + ", frame size " + msg.byteLength);
                    }

                    this.stats.calc(timestamps, msg.byteLength);
                } else {
                    logger.debug("Skipping frame = " + this.stats.frameCtr + " to " + this.nextFrameNo);
                }
            } catch(err) {
                logger.error("Error, msg " + err);
            }
        }.bind(this);

    }.bind(this);

    this.handleStatRequest = function(jsonMessage) {
        var sendOnlyAck = jsonMessage[XRM_SVR_STAT_REQUEST_ONLY_ACK];
        var serverFrameNumber = jsonMessage[XRM_SVR_STAT_REQUEST_SVR_FRNO];
        if (sendOnlyAck == false) {
            this.stats.sendStats(this.dataWS, serverFrameNumber, jsonMessage[XRM_MSG_SCREEN_ID]);
        } else {
            var obj = {};
            obj[XRM_MSG_TYPE] = XRM_MSGTYPE_CLNT_ACK_STAT;
            obj[XRM_MSG_SCREEN_ID] = this.screenID;
            obj[XRM_SVR_STAT_REQUEST_SVR_FRNO] = serverFrameNumber;
            this.dataWS.send(JSON.stringify(obj));
        }
    }.bind(this);

    this.handleSkip = function(jsonMessage) {
        logger.log("Received Skip Frame Message");
        this.nextFrameNo = jsonMessage[XRM_SVR_SKIP_NEXTFRAME];
        logger.log("Next frame : " + this.nextFrameNo + " currently at " + this.stats.frameCtr);
    }.bind(this);

    this.handleVideoSourceParam = function(jsonMessage, nativeDecoderEnabled, serverFeatures, webGLSupported) {
        logger.log("Received VideoSource Parameter");

        var canvas_w = 0;
        var canvas_h = 0;

        var newServerScreenWidth = 0;
        var newServerScreenHeight = 0;

        var serverFrameRate;    // 24;
        var serverQuality;      // 21;

        var screenID = 0;
        var encoding = "";

        screenID = parseInt(jsonMessage[XRM_MSG_SCREEN_ID]);
        serverFrameRate = parseInt(jsonMessage[XRM_SVR_VS_PARAM_FRAMERATE]);
        encoding = jsonMessage[XRM_VCONF];
        this.currentEncoding = encoding;

        StatsAvgWindow = serverFrameRate * 3;
        StatsStdWindow = serverFrameRate * 4;
        StatsAggrWindow = serverFrameRate * 8;

        this.stats.targetFrameTime = 1000 / serverFrameRate;

        this.stats.reinit();

        logger.log("Target Frame time: " + this.stats.targetFrameTime);

        this.lastSeqNoReported = 0;

        serverQuality = parseInt(jsonMessage[XRM_SVR_VS_PARAM_QUALITY]);
        this.stats.configNo = parseInt(jsonMessage[XRM_SVR_VS_PARAM_CONFIG_NO]);
        this.nextFrameNo = parseInt(jsonMessage[XRM_SVR_VS_PARAM_NEXTFRAME]);

        canvas_w = parseInt(jsonMessage[XRM_SVR_VS_PARAM_WIDTH]);
        canvas_h = parseInt(jsonMessage[XRM_SVR_VS_PARAM_HEIGHT]);

        newServerScreenWidth = parseInt(jsonMessage[XRM_SVR_VS_PARAM_SERVER_WIDTH]);
        newServerScreenHeight = parseInt(jsonMessage[XRM_SVR_VS_PARAM_SERVER_HEIGHT]);

        logger.debug("Checking canvas");

        var streamScaled = false;

        if (canvas_w !== newServerScreenWidth || canvas_h !== newServerScreenHeight) {
            streamScaled = true;
        }

        if (this.canvas.width !== canvas_w || this.canvas.height !== canvas_h) {
            this.canvas.width = canvas_w;
            this.canvas.height = canvas_h;

            logger.log("Updated canvas size to " + this.canvas.width + "x" + this.canvas.height);

            var $canvas = $("#canvas");

            if (nativeDecoderEnabled === false) {
                logger.debug("Initialize H.264 decoder with " + this.canvas.width + "x" + this.canvas.height);

                var avcPlayerStarted = (this.avcPlayer !== null);

                // initialize H.264 decoder
                if (avcPlayerStarted) {
                    this.avcPlayer.setup(this.canvas, rendermode);
                } else {
                    this.avcPlayer = new AvcPlayer(this.canvas, rendermode);
                }
            } else {
                // Initialize the HTML5 canvas so we can see the
                // native decoder's output.
                try {
                    var gl = this.canvas.getContext("experimental-webgl") || this.canvas.getContext("webgl");
                } catch(err) {
                    logger.error("Error: canvas does not have a webgl context");
                }
                if (gl) {
                    // Does not interfere with rest of page style.
                    gl.viewport(0, 0, this.canvas.width, this.canvas.height);
                } else {
                    // Allows output to canvas to be seen, but body
                    // element's background will never be drawn.
                    enableTransparency();
                }
            }
        } else {
            logger.debug("Canvas unchanged " + this.canvas.width + "x" + this.canvas.height);
        }

        // Detected screen size change on the server side
        if (this.serverScreenWidth !== newServerScreenWidth || this.serverScreenHeight !== newServerScreenHeight) {
            this.serverScreenWidth = newServerScreenWidth;
            this.serverScreenHeight = newServerScreenHeight;

            // Need to catch case, when screen size is "only" adjusted to mod 16 screen resolution
            var serverScreenWidth_mod = newServerScreenWidth - (newServerScreenWidth % 16);
            var serverScreenHeight_mod = newServerScreenHeight - (newServerScreenHeight % 4);

            var text = null;
            if ((newServerScreenWidth !== canvas_w || newServerScreenHeight !== canvas_h) &&
                (serverScreenWidth_mod !== canvas_w || serverScreenHeight_mod !== canvas_h)) {

                text = "<strong>Video downscaled from " + newServerScreenWidth + "x" + newServerScreenHeight + " to " + canvas_w + "x" + canvas_h + ".</strong>  The remote screen resolution exceeds the recommended limits set by your system administrator.";

            } else if ((newServerScreenWidth !== canvas_w || newServerScreenHeight !== canvas_h) &&
                       (serverScreenWidth_mod === canvas_w || serverScreenHeight_mod === canvas_h)) {

                text = "<strong>Video downscaled from " + newServerScreenWidth + "x" + newServerScreenHeight + " to " + canvas_w + "x" + canvas_h + " to meet encoder requirements.</strong>";

            }

            if (text !== null) {
                $("#system-alert-downscale").alert("close");
                var systemAlert = new SystemAlert($("#system-alert-container"), text, "info", "system-alert-downscale");
                systemAlert.slideDown();
                setAutoHideSystemAlert("#system-alert-downscale", 6000);
                $(".downscale-widget").show();
                $("#downscale-text").text("Video downscaled: " + newServerScreenWidth + "x" + newServerScreenHeight + " to " + canvas_w + "x" + canvas_h);
            } else {
                $("#system-alert-downscale").alert("close");
                $(".downscale-widget").hide();
                $("#downscale-text").text("Video is downscaled");
            }
        }
        if (serverFeatures & XRM_SVR_PARAM_FEATURE_LOSSLESS) {
            var current = this.setVideoQualitySlider(encoding);

            if (streamScaled && current.isScalable && !(serverFeatures & XRM_SVR_PARAM_FEATURE_LOSSLESS_SCALED)) {
                logger.log("Disconnecting lossless client on scaled stream");
                document.getElementById("signout").click();
                return;
            }
        }
        if (serverFeatures & XRM_SVR_PARAM_FEATURE_LOSSLESS) {
            if (streamScaled && !(serverFeatures & XRM_SVR_PARAM_FEATURE_LOSSLESS_SCALED)) {
                $(".video-quality-widget").hide();
            } else {
                $(".video-quality-widget").show();
            }
        } else {
            $(".video-quality-widget").hide();
        }

    }.bind(this);

    // Only sets the widget value.  Does not talk to server.
    this.setVideoQualitySlider = function(encoding) {
        var current = null;
        for (var i = 0; i < VIDEO_QUALITY_SETTINGS.length; i++) {
            if (encoding == VIDEO_QUALITY_SETTINGS[i].code) {
                current = VIDEO_QUALITY_SETTINGS[i];
                $("#video-quality-slider").bootstrapSlider("setValue", i);
                break;
            }
        }
        if (current == null) {
            logger.error("Error: Unexpected video encoding: " + encoding + ".  Going with default.");
            current = VIDEO_QUALITY_SETTINGS[0];
        }
        return current;
    }.bind(this);

    this.handleGuestsPaused = function(guestVideoPaused) {
        var $va = $("#screenshare-control-icon");

        if (guestVideoPaused) {
            $va.removeClass("glyphicon-pause");
            $va.addClass("glyphicon-play");
            $("#screenshare-control-label").text("Resume Guests");
        } else {
            $va.removeClass("glyphicon-play");
            $va.addClass("glyphicon-pause");
            $("#screenshare-control-label").text("Pause Guests");
        }
    }.bind(this);

    this.handleVideoPaused = function(isPaused) {
        if (isPaused)
            $("#overlay").show();
        else
            $("#overlay").hide();
    }.bind(this);

    // Find the right method, call on correct element
    function toggleFullscreen(elem) {
        elem = elem || document.documentElement;

        if (!document.fullscreenElement && !document.mozFullScreenElement &&
            !document.webkitFullscreenElement && !document.msFullscreenElement) {

            if (nativeClientEnabled === true && webGLSupported === false) {
                $("html").addClass("transparent-background");
                $("html").addClass("overflow-hidden");
            }

            if (elem.requestFullscreen) {
                elem.requestFullscreen();
            } else if (elem.msRequestFullscreen) {
                elem.msRequestFullscreen();
            } else if (elem.mozRequestFullScreen) {
                elem.mozRequestFullScreen();
            } else if (elem.webkitRequestFullscreen) {
                elem.webkitRequestFullscreen(Element.ALLOW_KEYBOARD_INPUT);
            }
        } else {
            if (nativeClientEnabled === true && webGLSupported === false) {
                $("html").removeClass("transparent-background");
                $("html").removeClass("overflow-hidden");
            }
            if (document.exitFullscreen) {
                document.exitFullscreen();
            } else if (document.msExitFullscreen) {
                document.msExitFullscreen();
            } else if (document.mozCancelFullScreen) {
                document.mozCancelFullScreen();
            } else if (document.webkitExitFullscreen) {
                document.webkitExitFullscreen();
            }
        }
    }

    document.addEventListener('fullscreenchange', fullscreenChangeHandler);
    document.addEventListener('webkitfullscreenchange', fullscreenChangeHandler);
    document.addEventListener('mozfullscreenchange', fullscreenChangeHandler);
    document.addEventListener('MSFullscreenChange', fullscreenChangeHandler);

    function fullscreenChangeHandler() {
        if (!document.fullscreenElement && !document.webkitIsFullScreen && !document.mozFullScreen && !document.msFullscreenElement) {
            $("#fullscreen > span.far").removeClass("fa-toggle-on");
            $("#fullscreen > span.far").addClass("fa-toggle-off");
        } else {
            $("#fullscreen > span.far").removeClass("fa-toggle-off");
            $("#fullscreen > span.far").addClass("fa-toggle-on");
        }
    }

    /* set canvas in fullscreen when button is pressed */
    document.getElementById("fullscreen").onclick = function(e) {
        toggleFullscreen(document.getElementById("main"));
    }.bind(this);

    /**
     * Setup the video quality slider.
     **/
    var myVideoQualitySlider = $("#video-quality-slider").bootstrapSlider({
        ticks: [0, 1, 2],
        ticks_labels: ["normal", "visually lossless", "lossless"],
        min: 0,
        max: 2,
        step: 1,
        value: 0,
        tooltip: "hide",
        useCurrentValue: true
    });

    myVideoQualitySlider.on("slideStop", function() {
        var val = myVideoQualitySlider.bootstrapSlider("getValue");
        this.sendEncodingChangeRequest(VIDEO_QUALITY_SETTINGS[val].code);
    }.bind(this));

    /**
     * bootstrapSlider doesn't layout correctly when dropdown isn't showing.
     **/
    $(".dropdown").on("shown.bs.dropdown", function() {
        myVideoQualitySlider.bootstrapSlider("relayout");
    }.bind(this));

}
